# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
from pathlib import Path
from typing import Dict, List, Optional

import torch
from prettytable import MARKDOWN, PrettyTable
from torch.optim import Adam

from dan.ocr.decoder import GlobalHTADecoder
from dan.ocr.encoder import FCN_Encoder
from dan.ocr.transforms import Preprocessing

METRICS_TABLE_HEADER = {
    "cer": "CER (HTR-NER)",
    "cer_no_token": "CER (HTR)",
    "wer": "WER (HTR-NER)",
    "wer_no_token": "WER (HTR)",
    "wer_no_punct": "WER (HTR no punct)",
    "ner": "NER",
}
REVERSE_HEADER = {column: metric for metric, column in METRICS_TABLE_HEADER.items()}


def update_config(config: dict):
    """
    Complete the fields that are not JSON serializable.
    """

    # .dataset.datasets cast all values to Path
    config["dataset"]["datasets"] = {
        name: Path(path) for name, path in config["dataset"]["datasets"].items()
    }

    # .model.encoder.class = FCN_ENCODER
    config["model"]["encoder"]["class"] = FCN_Encoder

    # .model.decoder.class = GlobalHTADecoder
    config["model"]["decoder"]["class"] = GlobalHTADecoder

    # .model.lm.path to Path
    if config["model"].get("lm", {}).get("path"):
        config["model"]["lm"]["path"] = Path(config["model"]["lm"]["path"])

    # Update preprocessing type
    for prepro in config["training"]["data"]["preprocessings"]:
        prepro["type"] = Preprocessing(prepro["type"])

    # .training.output_folder to Path
    config["training"]["output_folder"] = Path(config["training"]["output_folder"])

    if config["training"]["transfer_learning"]:
        # .training.transfer_learning.encoder[1]
        config["training"]["transfer_learning"]["encoder"][1] = Path(
            config["training"]["transfer_learning"]["encoder"][1]
        )

        # .training.transfer_learning.decoder[1]
        config["training"]["transfer_learning"]["decoder"][1] = Path(
            config["training"]["transfer_learning"]["decoder"][1]
        )

    # Parse optimizers
    for optimizer_setup in config["training"]["optimizers"].values():
        # Only supported optimizer is Adam
        optimizer_setup["class"] = Adam

    # set nb_gpu if not present
    if config["training"]["device"]["nb_gpu"] is None:
        config["training"]["device"]["nb_gpu"] = torch.cuda.device_count()


def create_metrics_table(metrics: List[str]) -> PrettyTable:
    """
    Create a Markdown table to display metrics in (CER, WER, NER, etc)
    for each evaluated split.
    """
    table = PrettyTable(
        field_names=["Split"]
        + [title for metric, title in METRICS_TABLE_HEADER.items() if metric in metrics]
    )
    table.set_style(MARKDOWN)

    return table


def add_metrics_table_row(
    table: PrettyTable, split: str, metrics: Optional[Dict[str, int | float]]
) -> PrettyTable:
    """
    Add a row to an existing metrics Markdown table for the currently evaluated split.
    To create such table please refer to
    [create_metrics_table][dan.ocr.utils.create_metrics_table] function.
    """
    row = [split]
    for column in table.field_names:
        if column not in REVERSE_HEADER:
            continue

        metric_name = REVERSE_HEADER[column]
        metric_value = metrics.get(metric_name)
        row.append(round(metric_value * 100, 2) if metric_value is not None else "−")

    table.add_row(row)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Extract dataset from Arkindex using a corpus export.
"""

import argparse
import pathlib
from uuid import UUID

from dan.datasets.extract.arkindex import run

MANUAL_SOURCE = "manual"


def parse_source(source) -> str | bool:
    if source == MANUAL_SOURCE:
        return False

    try:
        UUID(source)
    except ValueError:
        raise argparse.ArgumentTypeError(f"`{source}` is not a valid UUID.")

    return source


def validate_char(char):
    if len(char) != 1:
        raise argparse.ArgumentTypeError(
            f"`{char}` (of length {len(char)}) is not a valid character. Must be a string of length 1."
        )

    return char


def add_extract_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "extract",
        description=__doc__,
        help=__doc__,
    )

    # Required arguments.
    parser.add_argument(
        "database",
        type=pathlib.Path,
        help="Path where the data were exported from Arkindex.",
    )
    parser.add_argument(
        "--dataset-id",
        nargs="+",
        type=UUID,
        help="ID of the dataset to extract from Arkindex.",
        required=True,
        dest="dataset_ids",
    )
    parser.add_argument(
        "--element-type",
        nargs="+",
        type=str,
        help="Type of elements to retrieve.",
        required=True,
    )
    parser.add_argument(
        "--output",
        type=pathlib.Path,
        help="Path where the data will be generated.",
        required=True,
    )

    # Optional arguments.
    parser.add_argument(
        "--entity-separators",
        type=validate_char,
        nargs="+",
        help="""
            Removes all text that does not appear in an entity or in the list of given ordered characters.
            If several separators follow each other, keep only the first to appear in the list.
            Do not give any arguments to keep the whole text.
        """,
    )
    parser.add_argument(
        "--unknown-token",
        type=str,
        help="Token to use to replace character in the validation/test sets that is not included in the training set.",
        default="⁇",
    )
    parser.add_argument(
        "--tokens",
        type=pathlib.Path,
        help="Mapping between starting tokens and end tokens to extract text with their entities.",
        required=False,
    )

    parser.add_argument(
        "--transcription-worker-versions",
        type=parse_source,
        nargs="+",
        help=f"Filter transcriptions by worker_version. Use {MANUAL_SOURCE} for manual filtering.",
        default=[],
    )
    parser.add_argument(
        "--entity-worker-versions",
        type=parse_source,
        nargs="+",
        help=f"Filter transcriptions entities by worker_version. Use {MANUAL_SOURCE} for manual filtering.",
        default=[],
    )
    parser.add_argument(
        "--transcription-worker-runs",
        type=parse_source,
        nargs="+",
        help=f"Filter transcriptions by worker_run. Use {MANUAL_SOURCE} for manual filtering.",
        default=[],
    )
    parser.add_argument(
        "--entity-worker-runs",
        type=parse_source,
        nargs="+",
        help=f"Filter transcriptions entities by worker_run. Use {MANUAL_SOURCE} for manual filtering.",
        default=[],
    )

    parser.add_argument(
        "--subword-vocab-size",
        type=int,
        default=1000,
        help="Size of the vocabulary to train the sentencepiece subword tokenizer needed for language model.",
    )

    # Formatting arguments
    parser.add_argument(
        "--keep-spaces",
        action="store_true",
        help="Do not remove beginning, ending and consecutive spaces in transcriptions.",
    )

    parser.add_argument(
        "--allow-empty",
        action="store_true",
        help="Also extract data from element with no transcription.",
    )

    parser.set_defaults(func=run)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Extract entities from Arkindex using a corpus export.
"""
from pathlib import Path

from dan.datasets.entities.extract import run


def add_entities_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "entities",
        description=__doc__,
        help=__doc__,
    )
    parser.add_argument(
        "database",
        type=Path,
        help="Path where the data were exported from Arkindex.",
    )
    parser.add_argument(
        "--output-file",
        type=Path,
        default=Path("entities.yml"),
        required=False,
        help="Path to a YAML file to save the extracted entities.",
    )
    parser.set_defaults(func=run)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import copy
import json

import numpy as np
from torch.utils.data import Dataset

from dan.datasets.utils import natural_sort
from dan.utils import read_image, token_to_ind


class OCRDataset(Dataset):
    """
    Dataset class to handle dataset loading
    """

    def __init__(
        self,
        set_name,
        paths_and_sets,
        charset,
        tokens,
        preprocessing_transforms,
        augmentation_transforms,
        load_in_memory=False,
        mean=None,
        std=None,
    ):
        self.set_name = set_name
        self.charset = charset
        self.tokens = tokens
        self.load_in_memory = load_in_memory
        self.mean = mean
        self.std = std

        # Pre-processing, augmentation
        self.preprocessing_transforms = preprocessing_transforms
        self.augmentation_transforms = augmentation_transforms

        # Factor to reduce the height and width of the feature vector before feeding the decoder.
        self.reduce_dims_factor = np.array([32, 8, 1])

        # Load samples and preprocess images if load_in_memory is True
        self.samples = self.load_samples(paths_and_sets)

        # Curriculum config
        self.curriculum_config = None

    def __len__(self):
        """
        Return the dataset size
        """
        return len(self.samples)

    def __getitem__(self, idx):
        """
        Return an item from the dataset (image and label)
        """
        # Load preprocessed image
        sample = copy.deepcopy(self.samples[idx])
        if not self.load_in_memory:
            sample["img"] = self.get_sample_img(idx)

        # Convert to numpy
        sample["img"] = np.array(sample["img"])

        # Apply data augmentation
        if self.augmentation_transforms:
            sample["img"] = self.augmentation_transforms(image=sample["img"])["image"]

        # Image normalization
        sample["img"] = (sample["img"] - self.mean) / self.std

        # Get final height and width
        sample["img_reduced_shape"], sample["img_position"] = self.compute_final_size(
            sample["img"]
        )

        # Convert label into tokens
        sample["token_label"], sample["label_len"] = self.convert_sample_label(
            sample["label"]
        )
        return sample

    def load_samples(self, paths_and_sets):
        """
        Load images and labels
        """
        samples = list()

        for path_and_set in paths_and_sets:
            path = path_and_set["path"]
            gt_per_set = json.loads((path / "labels.json").read_text())
            set_name = path_and_set["set_name"]
            gt = gt_per_set[set_name]
            for filename in natural_sort(gt):
                filepath = path / filename
                samples.append(
                    {
                        "name": filepath.name,
                        "label": gt[filename],
                        "path": filepath.resolve(),
                    }
                )
                if self.load_in_memory:
                    samples[-1]["img"] = self.preprocessing_transforms(
                        read_image(str(filepath))
                    )
        return samples

    def get_sample_img(self, i):
        """
        Get image by index
        """
        if self.load_in_memory:
            return self.samples[i]["img"]

        image_path = str(self.samples[i]["path"])
        return self.preprocessing_transforms(read_image(image_path))

    def compute_std_mean(self):
        """
        Compute cumulated variance and mean of whole dataset
        """
        if self.mean is not None and self.std is not None:
            return self.mean, self.std

        total = np.zeros((3,))
        diff = np.zeros((3,))
        nb_pixels = 0
        for metric in ["mean", "std"]:
            for ind in range(len(self.samples)):
                img = np.array(self.get_sample_img(ind))
                if metric == "mean":
                    total += np.sum(img, axis=(0, 1))
                    nb_pixels += np.prod(img.shape[:2])
                elif metric == "std":
                    diff += [
                        np.sum((img[:, :, k] - self.mean[k]) ** 2) for k in range(3)
                    ]
            if metric == "mean":
                self.mean = total / nb_pixels
            elif metric == "std":
                self.std = np.sqrt(diff / nb_pixels)
        return self.mean, self.std

    def compute_final_size(self, img):
        """
        Compute the final image size and position after feature extraction
        """
        image_reduced_shape = np.ceil(img.shape / self.reduce_dims_factor).astype(int)

        if self.set_name == "train":
            image_reduced_shape = [max(1, t) for t in image_reduced_shape]

        image_position = [
            [0, img.shape[0]],
            [0, img.shape[1]],
        ]
        return image_reduced_shape, image_position

    def convert_sample_label(self, label):
        """
        Tokenize the label and return its length
        """
        token_label = token_to_ind(self.charset, label)
        token_label.append(self.tokens["end"])
        label_len = len(token_label)
        token_label.insert(0, self.tokens["start"])
        return token_label, label_len



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import pytest
from mdutils.mdutils import MdUtils

from dan.datasets.analyze import read_yaml
from dan.datasets.analyze.statistics import Statistics
from tests.conftest import FIXTURES


@pytest.fixture
def image_statistics():
    return MdUtils(file_name="").read_md_file(str(FIXTURES / "analyze" / "images"))


@pytest.fixture
def labels_statistics():
    return MdUtils(file_name="").read_md_file(str(FIXTURES / "analyze" / "labels"))


@pytest.fixture
def ner_statistics():
    return MdUtils(file_name="").read_md_file(str(FIXTURES / "analyze" / "ner"))


@pytest.fixture
def full_statistics():
    return MdUtils(file_name="").read_md_file(str(FIXTURES / "analyze" / "stats"))


def test_display_image_statistics(image_statistics, tmp_path):
    stats = Statistics(filename=tmp_path)
    stats.create_image_statistics(
        images=[
            "tests/data/training/training_dataset/images/0a34e13a-4ab0-4a91-8d7c-b1d8fee32628.png",
            "tests/data/training/training_dataset/images/0a70e14f-feda-4607-989c-36cf581ddff5.png",
            "tests/data/training/training_dataset/images/0a576062-303c-4893-a729-c09c92865d31.png",
            "tests/data/training/training_dataset/images/0b2457c8-81f1-4600-84d9-f8bf2822a991.png",
            "tests/data/training/training_dataset/images/fb3edb59-3678-49f8-8e16-8e32e3b0f051.png",
            "tests/data/training/training_dataset/images/fe498de2-ece4-4fbe-8b53-edfce1b820f0.png",
        ]
    )
    assert stats.document.get_md_text() == image_statistics


def test_display_label_statistics(labels_statistics, tmp_path):
    filename = tmp_path / "labels.md"
    stats = Statistics(filename=str(filename))
    stats.create_label_statistics(
        labels=[
            "Teklia’s expertise is to develop document analysis\nand processing solutions using, among other things,\nOCR technology.",
            "Our software combines image analysis, printed and\nhandwritten text recognition, text segmentation with\na document classification and indexation system.",
            "Our objective is to deliver to our clients an automated\ndocument processing tool easy-to-use and adapted\nto their needs.",
            "With the same state of mind, we developed additional solutions to\nenhance both security and business-process.",
        ]
    )
    assert stats.document.get_md_text() == labels_statistics


def test_display_ner_statistics(ner_statistics, tmp_path):
    tokens = read_yaml(FIXTURES / "training" / "training_dataset" / "tokens.yaml")
    stats = Statistics(filename=tmp_path)
    stats.create_ner_statistics(
        labels=[
            "ⓈDayon ⒻFernand Ⓐ6\nⓈDayen ⒻMaurice Ⓐ2\nⓈTottelier ⒻJean Baptiste Ⓐ59",
            "ⓈPeryro ⒻEtienne Ⓐ33\nⓈJeannot ⒻCaroline Ⓐ24\nⓈMouline ⒻPierre Ⓐ32",
        ],
        ner_tokens=tokens,
    )
    assert stats.document.get_md_text() == ner_statistics


def test_run(full_statistics, tmp_path):
    output_file = tmp_path / "stats.md"
    stats = Statistics(filename=str(output_file))
    stats.run(
        labels_path=FIXTURES / "training" / "training_dataset" / "labels.json",
        tokens=read_yaml(FIXTURES / "training" / "training_dataset" / "tokens.yaml"),
    )
    assert output_file.read_text() == full_statistics



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import itertools
import logging
import operator
import re
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, Iterator, List

import sentencepiece as spm
from lxml.etree import Element, SubElement, tostring
from nltk import wordpunct_tokenize

from arkindex_export import TranscriptionEntity
from dan.utils import EntityType, LMTokenMapping

logger = logging.getLogger(__name__)

# Replace \t with regular space and consecutive spaces
TRIM_SPACE_REGEX = re.compile(r"[\t ]+")
TRIM_RETURN_REGEX = re.compile(r"[\r\n]+")

# Remove invalid characters to build valid XML tag name
SLUG_PATTERN = re.compile(r"[\W]+")

# Some characters are encoded in XML but we don't want them encoded in the end
ENCODING_MAP = {
    "&#13;": "\r",
    "&lt;": "<",
    "&gt;": ">",
    "&amp;": "&",
}


def normalize_linebreaks(text: str) -> str:
    """
    Remove begin/ending linebreaks.
    Replace \r with regular linebreak and consecutive linebreaks.
    :param text: Text to normalize.
    """
    return TRIM_RETURN_REGEX.sub("\n", text.strip())


def normalize_spaces(text: str) -> str:
    """
    Remove begin/ending spaces.
    Replace \t with regular space and consecutive spaces.
    :param text: Text to normalize.
    """
    return TRIM_SPACE_REGEX.sub(" ", text.strip())


def get_vocabulary(tokenized_text: List[str]) -> set[str]:
    """
    Compute set of vocabulary from tokenzied text.
    :param tokenized_text: List of tokenized text.
    """
    return sorted(set([token for doc in tokenized_text for token in doc.split()]))


@dataclass
class Tokenizer:
    """
    A multi-level tokenizer (char, subword, word), where the subword tokenizer is trained using sentencepiece.
    :param training_corpus: List of training text.
    :param outdir: Path to save the subword tokenizer.
    :param mapping: Mapping between displayed and encoded versions of special characters.
    :param tokens: Start and end tokens used to represent named entities.
    :param subword_vocab_size: Size of the vocabulary size to use to train the subword tokenizer.
    """

    training_corpus: List[str]
    charset: List[str]
    unknown_token: str
    outdir: Path
    mapping: LMTokenMapping
    tokens: EntityType | None = None
    subword_vocab_size: int = 1000
    sentencepiece_model: spm.SentencePieceProcessor = field(init=False)

    @property
    def prefix(self):
        return self.outdir / "subword_tokenizer"

    @property
    def ner_tokens(self) -> List[str] | Iterator[str]:
        if self.tokens is None:
            return []
        return itertools.chain(
            map(operator.attrgetter("start"), self.tokens.values()),
            filter(
                operator.truth, map(operator.attrgetter("end"), self.tokens.values())
            ),
        )

    @property
    def mapping_tokens(self) -> List[str]:
        return [token.encoded for token in self.mapping]

    @property
    def special_tokens(self) -> List[str]:
        return list(set(itertools.chain(self.mapping_tokens, self.ner_tokens)))

    def __post_init__(self) -> None:
        """
        Train a sentencepiece model on the training corpus.
        """
        # Write the corpus in a text file
        logger.info("Training a sentencepiece model for subword tokenization")
        with NamedTemporaryFile(dir=self.outdir, suffix=".txt", mode="w") as tmp:
            tmp.write("\n".join(self.training_corpus))
            tmp.flush()

            try:
                spm.SentencePieceTrainer.train(
                    input=tmp.name,
                    vocab_size=self.subword_vocab_size,
                    model_prefix=self.prefix,
                    user_defined_symbols=self.special_tokens,
                    minloglevel=1,
                )
            except Exception as e:
                logger.warning(
                    f"Failed to train a sentencepiece model for subword tokenization: {e} "
                    "Try again by editing the `--subword-vocab-size` parameter."
                )
                self.sentencepiece_model = None
                return

        # Load the model
        self.sentencepiece_model = spm.SentencePieceProcessor(
            model_file=str(self.prefix.with_suffix(".model"))
        )

    def subword_tokenize(self, text: str) -> str:
        """
        Tokenize into subwords. Sampling is disabled to ensure reproducibility.
        """
        tokens = self.sentencepiece_model.encode(text, out_type=str)
        return " ".join(map("".join, map(self.encode, tokens)))

    def word_tokenize(self, text: str) -> str:
        """
        Tokenize text into a string of space-separated words. Spaces (⎵) and NER tokens are considered as words.
        :param text: Text to be tokenized.
        """
        words = list(map("".join, map(self.encode, wordpunct_tokenize(text))))
        return " ".join(
            [
                word + f" {self.mapping.space.encoded}"
                if (i != len(words) - 1 and word not in self.ner_tokens)
                else word
                for i, word in enumerate(words)
            ]
        )

    def char_tokenize(self, text: str) -> str:
        """
        Tokenize text into a string of space-separated characters.
        :param text: Text to be tokenized.
        """
        return " ".join(
            [
                char if char in self.charset else self.unknown_token
                for char in self.encode(text)
            ]
        )

    def encode(self, text: List[str]) -> List[str]:
        """
        Encode special tokens.
        :param text: Text to be encoded.
        """
        return map(self.mapping.encode_token, text)


def slugify(text: str):
    """
    Replace invalid characters in text to underscores to use it as XML tag.
    """
    return SLUG_PATTERN.sub("_", text)


def get_translation_map(tokens: Dict[str, EntityType]) -> Dict[str, str] | None:
    if not tokens:
        return

    translation_map = {
        # Roots
        "<root>": "",
        "</root>": "",
    }
    # Tokens
    for entity_name, token_type in tokens.items():
        translation_map[f"<{slugify(entity_name)}>"] = token_type.start
        translation_map[f"</{slugify(entity_name)}>"] = token_type.end

    return translation_map


@dataclass
class XMLEntity:
    type: str
    name: str
    offset: int
    length: int
    worker_version: str
    worker_run: str
    children: List["XMLEntity"] = field(default_factory=list)

    @property
    def end(self) -> int:
        return self.offset + self.length

    def add_child(self, child: TranscriptionEntity):
        self.children.append(
            XMLEntity(
                type=child["type"],
                name=child["name"],
                offset=child["offset"] - self.offset,
                length=child["length"],
                worker_version=child["worker_version"],
                worker_run=child["worker_run"],
            )
        )

    def insert(self, parent: Element):
        e = SubElement(parent, slugify(self.type))

        if not self.children:
            # No children
            e.text = self.name
            return

        offset = 0
        for child in self.children:
            # Add text before entity
            portion_before = self.name[offset : child.offset]
            offset += len(portion_before)
            if len(e):
                e[-1].tail = portion_before
            else:
                e.text = portion_before
            child.insert(e)
            offset += child.length

        # Text after the last entity
        e[-1].tail = self.name[self.children[-1].end : self.end]


def entities_to_xml(
    text: str,
    predictions: List[TranscriptionEntity],
    entity_separators: List[str] | None = None,
) -> str:
    """Represent the transcription and its entities in XML format. Each entity will be exposed with an XML tag.
    Its type will be used to name the tag.

    :param text: The text of the transcription
    :param predictions: The list of entities linked to the transcription
    :param entity_separators: When provided, instead of adding the text between entities, add one separator encountered in this text. The order is kept when looking for separators. Defaults to None
    :return: The representation of the transcription in XML format
    """

    def _find_separator(transcription: str) -> str:
        """
        Find the first entity separator in the provided transcription.
        """
        for separator in entity_separators:
            if separator in transcription:
                return separator
        return ""

    def add_portion(entity_offset: int | None = None):
        """
        Add the portion of text between entities either:
        - after the last node, if there is one before
        - on this node

        If we remove the text between entities, we keep one of the separators provided. Order matters.
        """
        portion = text[offset:entity_offset]

        if entity_separators:
            # Remove the text except the first entity_separator encountered
            portion = _find_separator(portion)

        if len(root):
            root[-1].tail = portion
        else:
            root.text = portion

    entities = iter(predictions)

    # This will mark the ending position of the first-level of entities
    last_end = None
    parsed: List[XMLEntity] = []

    for entity in entities:
        # First entity is not inside any other
        # If offset is too high, no nestation
        if not last_end or entity["offset"] >= last_end:
            parsed.append(XMLEntity(**entity))
            last_end = entity["offset"] + entity["length"]
            continue

        # Nested entity
        parsed[-1].add_child(entity)

    # XML export
    offset = 0
    root = Element("root")

    for entity in parsed:
        add_portion(entity.offset)

        entity.insert(root)

        offset = entity.end

    # Add text after last entity
    add_portion()

    # Cleanup separators introduced when text was removed
    if entity_separators:
        characters = "".join(entity_separators)
        root.text = root.text.lstrip(characters)
        # Strip trailing spaces on last child
        root[-1].tail = root[-1].tail.rstrip(characters)

    encoded_transcription = tostring(root, encoding="utf-8").decode()
    for pattern, repl in ENCODING_MAP.items():
        encoded_transcription = encoded_transcription.replace(pattern, repl)
    return encoded_transcription



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import pytest

from dan.datasets.tokens.generate import LIMIT, OFFSET, get_token, run
from tests import FIXTURES

TOKENS_DATA_PATH = FIXTURES / "tokens"


def test_get_token():
    token_generator = get_token()

    tokens = []
    for _ in range(LIMIT - OFFSET):
        tokens.append(next(token_generator))

    assert tokens == [
        "Ⓐ",
        "Ⓑ",
        "Ⓒ",
        "Ⓓ",
        "Ⓔ",
        "Ⓕ",
        "Ⓖ",
        "Ⓗ",
        "Ⓘ",
        "Ⓙ",
        "Ⓚ",
        "Ⓛ",
        "Ⓜ",
        "Ⓝ",
        "Ⓞ",
        "Ⓟ",
        "Ⓠ",
        "Ⓡ",
        "Ⓢ",
        "Ⓣ",
        "Ⓤ",
        "Ⓥ",
        "Ⓦ",
        "Ⓧ",
        "Ⓨ",
        "Ⓩ",
        "ⓐ",
        "ⓑ",
        "ⓒ",
        "ⓓ",
        "ⓔ",
        "ⓕ",
        "ⓖ",
        "ⓗ",
        "ⓘ",
        "ⓙ",
        "ⓚ",
        "ⓛ",
        "ⓜ",
        "ⓝ",
        "ⓞ",
        "ⓟ",
        "ⓠ",
        "ⓡ",
        "ⓢ",
        "ⓣ",
        "ⓤ",
        "ⓥ",
        "ⓦ",
        "ⓧ",
        "ⓨ",
        "ⓩ",
        "⓪",
        "⓫",
        "⓬",
        "⓭",
        "⓮",
        "⓯",
        "⓰",
        "⓱",
        "⓲",
        "⓳",
        "⓴",
        "⓵",
        "⓶",
        "⓷",
        "⓸",
        "⓹",
        "⓺",
        "⓻",
        "⓼",
        "⓽",
        "⓾",
        "⓿",
    ]


@pytest.mark.parametrize(
    "end_tokens, expected_file",
    [
        (True, TOKENS_DATA_PATH / "end_tokens.yml"),
        (False, TOKENS_DATA_PATH / "no_end_tokens.yml"),
    ],
)
def test_tokens(end_tokens, expected_file, tmp_path):
    output_file = tmp_path / "tokens.yml"

    run(
        entities=FIXTURES / "entities.yml",
        end_tokens=end_tokens,
        output_file=output_file,
    )

    assert output_file.read_text() == expected_file.read_text()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import logging
import pickle
import random
from collections import defaultdict
from pathlib import Path
from typing import Dict, List
from uuid import UUID

from tqdm import tqdm

from arkindex_export import Dataset, DatasetElement, Element, open_database
from dan.datasets.extract.db import (
    get_dataset_elements,
    get_elements,
    get_transcription_entities,
    get_transcriptions,
)
from dan.datasets.extract.exceptions import (
    NoTranscriptionError,
    ProcessingError,
    UnknownTokenInText,
)
from dan.datasets.extract.utils import (
    Tokenizer,
    entities_to_xml,
    get_translation_map,
    get_vocabulary,
    normalize_linebreaks,
    normalize_spaces,
)
from dan.utils import LMTokenMapping, parse_tokens

LANGUAGE_DIR = "language_model"  # Subpath to the language model directory.

TRAIN_NAME = "train"
VAL_NAME = "val"
TEST_NAME = "test"
SPLIT_NAMES = [TRAIN_NAME, VAL_NAME, TEST_NAME]

logger = logging.getLogger(__name__)


class ArkindexExtractor:
    """
    Extract data from Arkindex
    """

    def __init__(
        self,
        dataset_ids: List[UUID] | None = None,
        element_type: List[str] = [],
        output: Path | None = None,
        entity_separators: List[str] = ["\n", " "],
        unknown_token: str = "⁇",
        tokens: Path | None = None,
        transcription_worker_versions: List[str | bool] = [],
        entity_worker_versions: List[str | bool] = [],
        transcription_worker_runs: List[str | bool] = [],
        entity_worker_runs: List[str | bool] = [],
        keep_spaces: bool = False,
        allow_empty: bool = False,
        subword_vocab_size: int = 1000,
    ) -> None:
        self.dataset_ids = dataset_ids
        self.element_type = element_type
        self.output = output
        self.entity_separators = entity_separators
        self.unknown_token = unknown_token
        self.tokens = parse_tokens(tokens) if tokens else {}
        self.transcription_worker_versions = transcription_worker_versions
        self.entity_worker_versions = entity_worker_versions
        self.transcription_worker_runs = transcription_worker_runs
        self.entity_worker_runs = entity_worker_runs
        self.allow_empty = allow_empty
        self.mapping = LMTokenMapping()
        self.keep_spaces = keep_spaces
        self.subword_vocab_size = subword_vocab_size

        self.data: Dict = defaultdict(dict)
        self.charset = set()
        self.language_corpus = defaultdict(list)
        self.language_tokens = []
        self.language_lexicon = defaultdict(list)

        # NER extraction
        self.translation_map: Dict[str, str] | None = get_translation_map(self.tokens)

    def translate(self, text: str):
        """
        Use translation map to replace XML tags to actual tokens
        """
        for pattern, repl in self.translation_map.items():
            text = text.replace(pattern, repl)
        return text

    def extract_transcription(self, element: Element):
        """
        Extract the element's transcription.
        If the entities are needed, they are added to the transcription using tokens.
        """
        transcriptions = get_transcriptions(
            element.id,
            self.transcription_worker_versions,
            self.transcription_worker_runs,
        )
        if len(transcriptions) == 0:
            if self.allow_empty:
                return ""
            raise NoTranscriptionError(element.id)

        transcription = random.choice(transcriptions)
        stripped_text = transcription.text.strip()

        if not self.tokens:
            return stripped_text

        entities = get_transcription_entities(
            transcription.id,
            self.entity_worker_versions,
            self.entity_worker_runs,
            supported_types=list(self.tokens),
        )

        if not entities.count():
            return stripped_text

        return self.translate(
            entities_to_xml(
                transcription.text, entities, entity_separators=self.entity_separators
            )
        )

    def format_text(self, text: str, charset: set | None = None):
        if not self.keep_spaces:
            text = normalize_spaces(text)
            text = normalize_linebreaks(text)

        # Replace unknown characters by the unknown token
        if charset is not None:
            unknown_charset = set(text) - charset
            text = text.translate(
                {
                    ord(unknown_char): self.unknown_token
                    for unknown_char in unknown_charset
                }
            )
        return text.strip()

    def process_element(self, dataset_parent: DatasetElement, element: Element):
        """
        Extract an element's data and save it to disk.
        The output path is directly related to the split of the element.
        """
        text = self.extract_transcription(element)

        if self.unknown_token in text:
            raise UnknownTokenInText(element_id=element.id)

        text = self.format_text(
            text,
            # Do not replace unknown characters in train split
            charset=self.charset if dataset_parent.set_name != TRAIN_NAME else None,
        )

        self.data[dataset_parent.set_name][element.id] = {
            "dataset_id": dataset_parent.dataset_id,
            "text": text,
            "image": {
                "iiif_url": element.image.url,
                "polygon": json.loads(element.polygon),
            },
        }

        self.charset = self.charset.union(set(text))

    def process_parent(self, pbar, dataset_parent: DatasetElement):
        """
        Extract data from a parent element.
        """
        parent = dataset_parent.element
        base_description = f"Extracting data from {parent.type} ({parent.id}) for split ({dataset_parent.set_name})"
        pbar.set_description(desc=base_description)
        if self.element_type == [parent.type]:
            try:
                self.process_element(dataset_parent, parent)
            except ProcessingError as e:
                logger.warning(f"Skipping {parent.id}: {str(e)}")
        # Extract children elements
        else:
            children = get_elements(
                parent.id,
                self.element_type,
            )

            nb_children = children.count()
            for idx, element in enumerate(children, start=1):
                # Update description to update the children processing progress
                pbar.set_description(desc=base_description + f" ({idx}/{nb_children})")
                try:
                    self.process_element(dataset_parent, element)
                except ProcessingError as e:
                    logger.warning(f"Skipping {element.id}: {str(e)}")

    def format_lm_files(self) -> None:
        """
        Convert charset to a LM-compatible charset. Ensure that special LM tokens do not appear in the charset.
        """
        logger.info("Preparing language resources")
        # Add unknown token to charset
        self.charset.add(self.unknown_token)

        # Build LM tokens
        for token in sorted(list(self.charset)):
            assert (
                token not in self.mapping.encode.values()
            ), f"Special token {token} is reserved for language modeling."
            self.language_tokens.append(
                self.mapping.encode[token]
            ) if token in self.mapping.encode else self.language_tokens.append(token)
        self.language_tokens.append(self.mapping.ctc.encoded)

        # Build LM corpus
        train_corpus = [
            values["text"].replace(
                self.mapping.linebreak.display, self.mapping.space.display
            )
            for values in self.data[TRAIN_NAME].values()
        ]

        tokenizer = Tokenizer(
            training_corpus=train_corpus,
            charset=self.language_tokens,
            unknown_token=self.unknown_token,
            outdir=self.output / "language_model",
            mapping=self.mapping,
            tokens=self.tokens,
            subword_vocab_size=self.subword_vocab_size,
        )

        if not tokenizer.sentencepiece_model:
            return

        for level, tokenize in (
            ("characters", tokenizer.char_tokenize),
            ("words", tokenizer.word_tokenize),
            ("subwords", tokenizer.subword_tokenize),
        ):
            self.language_corpus[level] = list(map(tokenize, train_corpus))

        # Build LM lexicon
        self.language_lexicon["characters"] = [
            f"{token} {token}" for token in self.language_tokens
        ]
        for level in ["words", "subwords"]:
            self.language_lexicon[level] = [
                f"{token} {tokenizer.char_tokenize(token)}"
                for token in get_vocabulary(self.language_corpus[level])
            ]

    def export(self):
        (self.output / "split.json").write_text(
            json.dumps(
                self.data,
                sort_keys=True,
                indent=4,
            )
        )
        for level in ["characters", "words", "subwords"]:
            (self.output / "language_model" / f"corpus_{level}.txt").write_text(
                "\n".join(self.language_corpus[level])
            )
            (self.output / "language_model" / f"lexicon_{level}.txt").write_text(
                "\n".join(self.language_lexicon[level])
            )
        (self.output / "language_model" / "tokens.txt").write_text(
            "\n".join(self.language_tokens)
        )
        (self.output / "charset.pkl").write_bytes(
            pickle.dumps(sorted(list(self.charset)))
        )

    def run(self):
        # Retrieve the Dataset and its splits from the cache
        for dataset_id in self.dataset_ids:
            dataset = Dataset.get_by_id(dataset_id)
            splits = dataset.sets.split(",")
            if not set(splits).issubset(set(SPLIT_NAMES)):
                logger.warning(
                    f'Dataset {dataset.name} ({dataset.id}) does not have "{TRAIN_NAME}", "{VAL_NAME}" and "{TEST_NAME}" steps'
                )
                continue

            # Extract the train set first to correctly build the `self.charset` variable
            splits.remove(TRAIN_NAME)
            splits.insert(0, TRAIN_NAME)

            # Iterate over the subsets to find the page images and labels.
            for split in splits:
                with tqdm(
                    get_dataset_elements(dataset, split),
                    desc=f"Extracting data from ({dataset_id}) for split ({split})",
                ) as pbar:
                    # Iterate over the pages to create splits at page level.
                    for parent in pbar:
                        self.process_parent(
                            pbar=pbar,
                            dataset_parent=parent,
                        )
                        # Progress bar updates
                        pbar.update()
                        pbar.refresh()

        if not self.data:
            raise Exception(
                "No data was extracted using the provided export database and parameters."
            )

        self.format_lm_files()
        self.export()


def run(
    database: Path,
    dataset_ids: List[UUID],
    element_type: List[str],
    output: Path,
    entity_separators: List[str],
    unknown_token: str,
    tokens: Path,
    transcription_worker_versions: List[str | bool],
    entity_worker_versions: List[str | bool],
    transcription_worker_runs: List[str | bool],
    entity_worker_runs: List[str | bool],
    keep_spaces: bool,
    allow_empty: bool,
    subword_vocab_size: int,
):
    assert database.exists(), f"No file found @ {database}"
    open_database(path=database)

    # Create directories
    Path(output, LANGUAGE_DIR).mkdir(parents=True, exist_ok=True)

    ArkindexExtractor(
        dataset_ids=dataset_ids,
        element_type=element_type,
        output=output,
        entity_separators=entity_separators,
        unknown_token=unknown_token,
        tokens=tokens,
        transcription_worker_versions=transcription_worker_versions,
        entity_worker_versions=entity_worker_versions,
        transcription_worker_runs=transcription_worker_runs,
        entity_worker_runs=entity_worker_runs,
        keep_spaces=keep_spaces,
        allow_empty=allow_empty,
        subword_vocab_size=subword_vocab_size,
    ).run()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
from typing import List

from arkindex_export import Image
from arkindex_export.models import (
    Dataset,
    DatasetElement,
    Element,
    Entity,
    EntityType,
    Transcription,
    TranscriptionEntity,
)
from arkindex_export.queries import list_children


def get_dataset_elements(
    dataset: Dataset,
    split: str,
):
    """
    Retrieve dataset elements in a specific split from an SQLite export of an Arkindex corpus
    """
    query = (
        DatasetElement.select()
        .join(Element)
        .join(Image, on=(DatasetElement.element.image == Image.id))
        .where(
            DatasetElement.dataset == dataset,
            DatasetElement.set_name == split,
        )
    )

    return query


def get_elements(
    parent_id: str,
    element_type: List[str],
):
    """
    Retrieve elements from an SQLite export of an Arkindex corpus
    """
    # And load all the elements found in the CTE
    query = (
        list_children(parent_id=parent_id)
        .join(Image)
        .where(Element.type.in_(element_type))
    )

    return query


def build_worker_version_filter(ArkindexModel, worker_versions: List[str | bool]):
    """
    `False` worker version means `manual` worker_version -> null field.
    """
    condition = None
    for worker_version in worker_versions:
        condition |= (
            ArkindexModel.worker_version == worker_version
            if worker_version
            else ArkindexModel.worker_version.is_null()
        )
    return condition


def build_worker_run_filter(ArkindexModel, worker_runs: List[str | bool]):
    """
    `False` worker run means `manual` worker_run -> null field.
    """
    condition = None
    for worker_run in worker_runs:
        condition |= (
            ArkindexModel.worker_run == worker_run
            if worker_run
            else ArkindexModel.worker_run.is_null()
        )
    return condition


def get_transcriptions(
    element_id: str,
    transcription_worker_versions: List[str | bool],
    transcription_worker_runs: List[str | bool],
) -> List[Transcription]:
    """
    Retrieve transcriptions from an SQLite export of an Arkindex corpus
    """
    query = Transcription.select(
        Transcription.id,
        Transcription.text,
        Transcription.worker_version,
        Transcription.worker_run,
    ).where((Transcription.element == element_id))

    if transcription_worker_versions:
        query = query.where(
            build_worker_version_filter(
                Transcription, worker_versions=transcription_worker_versions
            )
        )

    if transcription_worker_runs:
        query = query.where(
            build_worker_run_filter(
                Transcription, worker_runs=transcription_worker_runs
            )
        )

    return query


def get_transcription_entities(
    transcription_id: str,
    entity_worker_versions: List[str | bool],
    entity_worker_runs: List[str | bool],
    supported_types: List[str],
) -> List[TranscriptionEntity]:
    """
    Retrieve transcription entities from an SQLite export of an Arkindex corpus
    """
    query = (
        TranscriptionEntity.select(
            EntityType.name.alias("type"),
            Entity.name.alias("name"),
            TranscriptionEntity.offset,
            TranscriptionEntity.length,
            TranscriptionEntity.worker_version,
            TranscriptionEntity.worker_run,
        )
        .join(Entity, on=TranscriptionEntity.entity)
        .join(EntityType, on=Entity.type)
        .where(
            TranscriptionEntity.transcription == transcription_id,
            EntityType.name.in_(supported_types),
        )
    )

    if entity_worker_versions:
        query = query.where(
            build_worker_version_filter(
                TranscriptionEntity, worker_versions=entity_worker_versions
            )
        )

    if entity_worker_runs:
        query = query.where(
            build_worker_run_filter(TranscriptionEntity, worker_runs=entity_worker_runs)
        )

    return query.order_by(
        TranscriptionEntity.offset, TranscriptionEntity.length.desc()
    ).dicts()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import logging
import os
import random
from copy import deepcopy
from enum import Enum
from itertools import repeat
from pathlib import Path
from time import time
from typing import Dict, List, Tuple

import numpy as np
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
import yaml
from torch.cuda.amp import GradScaler, autocast
from torch.nn import CrossEntropyLoss
from torch.nn.init import kaiming_uniform_
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from dan.ocr.decoder import CTCLanguageDecoder, GlobalHTADecoder
from dan.ocr.encoder import FCN_Encoder
from dan.ocr.manager.metrics import Inference, MetricManager
from dan.ocr.manager.ocr import OCRDatasetManager
from dan.ocr.mlflow import MLFLOW_AVAILABLE, logging_metrics, logging_tags_metrics
from dan.ocr.schedulers import DropoutScheduler
from dan.utils import fix_ddp_layers_names, ind_to_token

if MLFLOW_AVAILABLE:
    import mlflow

logger = logging.getLogger(__name__)

MODEL_NAME_ENCODER = "encoder"
MODEL_NAME_DECODER = "decoder"
MODEL_NAMES = (MODEL_NAME_ENCODER, MODEL_NAME_DECODER)


class GenericTrainingManager:
    def __init__(self, params):
        self.type = None
        self.is_master = False
        self.params = params
        self.models = {}
        self.dataset = None
        self.dataset_name = list(self.params["dataset"]["datasets"].values())[0]
        self.paths = None
        self.latest_step = 0
        self.latest_epoch = -1

        self.scaler = None

        self.optimizers = dict()
        self.optimizers_named_params_by_group = dict()
        self.lr_schedulers = dict()
        self.best = None
        self.writer = None
        self.metric_manager = dict()

        self.device_params = self.params["training"]["device"]
        self.nb_gpu = (
            self.device_params["nb_gpu"]
            if self.device_params["use_ddp"]
            else torch.cuda.device_count()
        )
        # Number of worker that process. Set to the number of GPU available if we are using DDP. Otherwise set to 1.
        self.nb_workers = self.nb_gpu if self.device_params["use_ddp"] else 1
        self.tokens = self.params["dataset"].get("tokens")

        self.init_hardware_config()
        self.init_paths()
        self.load_dataset()

    @property
    def encoder(self) -> FCN_Encoder | None:
        return self.models.get(MODEL_NAME_ENCODER)

    @property
    def decoder(self) -> GlobalHTADecoder | None:
        return self.models.get(MODEL_NAME_DECODER)

    def init_paths(self):
        """
        Create output folders for results and checkpoints
        """
        output_path = self.params["training"]["output_folder"]
        checkpoints_path = output_path / "checkpoints"
        results_path = output_path / "results"

        # Create folder
        checkpoints_path.mkdir(parents=True, exist_ok=True)
        results_path.mkdir(exist_ok=True)

        self.paths: Dict[str, Path] = {
            "results": results_path,
            "checkpoints": checkpoints_path,
            "output_folder": output_path,
        }

    def load_dataset(self):
        """
        Load datasets, data samplers and data loaders
        """
        self.dataset = OCRDatasetManager(
            dataset_params=self.params["dataset"],
            training_params=self.params["training"],
            device=self.device,
        )
        self.dataset.load_datasets()
        self.dataset.load_ddp_samplers()
        self.dataset.load_dataloaders()

    def init_hardware_config(self):
        cuda_is_available = torch.cuda.is_available()

        # Debug mode
        if self.device_params["force"] not in [None, "cuda"] or not cuda_is_available:
            self.device_params["use_ddp"] = False
            self.device_params["use_amp"] = False

        # Manage Distributed Data Parallel & GPU usage
        self.manual_seed = self.params["training"].get("manual_seed", 1111)
        self.ddp_config = {
            "master": self.device_params["use_ddp"]
            and self.device_params["ddp_rank"] == 0,
            "address": self.device_params.get("ddp_addr", "localhost"),
            "port": self.device_params.get("ddp_addr", "11111"),
            "backend": self.device_params.get("ddp_backend", "nccl"),
            "rank": self.device_params["ddp_rank"],
        }
        self.is_master = self.ddp_config["master"] or not self.device_params["use_ddp"]
        if self.device_params["use_ddp"]:
            self.device = torch.device(self.ddp_config["rank"])
            self.device_params["ddp_rank"] = self.ddp_config["rank"]
            self.launch_ddp()
        else:
            self.device = torch.device(
                self.device_params["force"] or "cuda" if cuda_is_available else "cpu"
            )
        if self.device == torch.device("cpu"):
            self.params["model"]["device"] = "cpu"
        else:
            self.params["model"]["device"] = self.device.type
        # Print GPU info
        # global
        if self.ddp_config["master"] or not self.device_params["use_ddp"]:
            print("##################")
            print("Available GPUS: {}".format(self.nb_gpu))
            for i in range(self.nb_gpu):
                print(
                    "Rank {}: {} {}".format(
                        i,
                        torch.cuda.get_device_name(i),
                        torch.cuda.get_device_properties(i),
                    )
                )
            print("##################")
        # local
        print("Local GPU:")
        if self.device != torch.device("cpu"):
            print(
                "Rank {}: {} {}".format(
                    self.device_params["ddp_rank"],
                    torch.cuda.get_device_name(),
                    torch.cuda.get_device_properties(self.device),
                )
            )
        else:
            print("WORKING ON CPU !\n")
        print("##################")

    def load_model(self, reset_optimizer=False, strict=True):
        """
        Load model weights from scratch or from checkpoints
        """
        common_params = {
            "h_max": self.params["model"].get("h_max"),
            "w_max": self.params["model"].get("w_max"),
            "device": self.device,
            "vocab_size": self.params["model"]["vocab_size"],
        }
        # Instantiate encoder, decoder
        for model_name in MODEL_NAMES:
            params = self.params["model"][model_name]
            model_class = params.get("class")
            self.models[model_name] = model_class({**params, **common_params})
            self.models[model_name].to(self.device)  # To GPU or CPU
            # make the model compatible with Distributed Data Parallel if used
            if self.device_params["use_ddp"]:
                self.models[model_name] = DDP(
                    self.models[model_name],
                    [self.ddp_config["rank"]],
                    output_device=self.ddp_config["rank"],
                )

        # Instantiate LM decoder
        self.lm_decoder = None
        if self.params["model"].get("lm") and self.params["model"]["lm"]["weight"] > 0:
            logger.info(
                f"Decoding with a language model (weight={self.params['model']['lm']['weight']})."
            )
            # Check files
            model_path = self.params["model"]["lm"]["path"]
            assert model_path.is_file(), f"File {model_path} not found"
            base_path = model_path.parent
            lexicon_path = base_path / "lexicon.txt"
            assert lexicon_path.is_file(), f"File {lexicon_path} not found"
            tokens_path = base_path / "tokens.txt"
            assert tokens_path.is_file(), f"File {tokens_path} not found"
            # Load LM decoder
            self.lm_decoder = CTCLanguageDecoder(
                language_model_path=str(model_path),
                lexicon_path=str(lexicon_path),
                tokens_path=str(tokens_path),
                language_model_weight=self.params["model"]["lm"]["weight"],
            )

        # Handle curriculum dropout
        self.dropout_scheduler = DropoutScheduler(self.models)

        self.scaler = GradScaler(enabled=self.device_params["use_amp"])

        # Check if checkpoint exists
        checkpoint = self.get_checkpoint()
        if checkpoint is not None:
            self.load_existing_model(checkpoint, strict=strict)
        else:
            self.init_new_model()

        self.load_optimizers(checkpoint, reset_optimizer=reset_optimizer)

        if self.is_master:
            print("LOADED EPOCH: {}\n".format(self.latest_epoch), flush=True)

    def get_checkpoint(self):
        """
        Seek if checkpoint exist, return None otherwise
        """
        if self.params["training"]["load_epoch"] in ("best", "last"):
            for filename in self.paths["checkpoints"].iterdir():
                if self.params["training"]["load_epoch"] in filename.name:
                    return torch.load(
                        filename,
                        map_location=self.device,
                    )
        return None

    def load_existing_model(self, checkpoint, strict=True):
        """
        Load information and weights from previous training
        """
        self.load_save_info(checkpoint)
        self.latest_epoch = checkpoint["epoch"]
        if "step" in checkpoint:
            self.latest_step = checkpoint["step"]
        self.best = checkpoint["best"]
        if "scaler_state_dict" in checkpoint:
            self.scaler.load_state_dict(checkpoint["scaler_state_dict"])
        if "dropout_scheduler_step" in checkpoint:
            self.dropout_scheduler.resume(checkpoint["dropout_scheduler_step"])
        # Load model weights from past training
        for model_name in self.models:
            # Transform to DDP/from DDP model
            checkpoint[f"{model_name}_state_dict"] = fix_ddp_layers_names(
                checkpoint[f"{model_name}_state_dict"],
                self.device_params["use_ddp"],
            )

            self.models[model_name].load_state_dict(
                checkpoint[f"{model_name}_state_dict"], strict=strict
            )

    def init_new_model(self):
        """
        Initialize model
        """
        # Specific weights initialization if exists
        for model_name in self.models:
            try:
                self.models[model_name].init_weights()
            except Exception:
                pass

        # Handle transfer learning instructions
        if self.params["training"]["transfer_learning"]:
            # Iterates over models
            for model_name in self.params["training"]["transfer_learning"]:
                state_dict_name, path, learnable, strict = self.params["training"][
                    "transfer_learning"
                ][model_name]

                # Loading pretrained weights file
                checkpoint = torch.load(path, map_location=self.device)
                # Transform to DDP/from DDP model
                checkpoint[f"{model_name}_state_dict"] = fix_ddp_layers_names(
                    checkpoint[f"{model_name}_state_dict"],
                    self.device_params["use_ddp"],
                )

                try:
                    # Load pretrained weights for model
                    self.models[model_name].load_state_dict(
                        checkpoint["{}_state_dict".format(state_dict_name)],
                        strict=strict,
                    )
                    print(
                        "transferred weights for {}".format(state_dict_name), flush=True
                    )
                except RuntimeError as e:
                    print(e, flush=True)
                    # if error, try to load each parts of the model (useful if only few layers are different)
                    for key in checkpoint[
                        "{}_state_dict".format(state_dict_name)
                    ].keys():
                        try:
                            # for pre-training of decision layer
                            if (
                                "end_conv" in key
                                and "transfered_charset" in self.params["model"]
                                and self.params["model"]["transfered_charset"]
                            ):
                                self.adapt_decision_layer_to_old_charset(
                                    model_name, key, checkpoint, state_dict_name
                                )
                            else:
                                self.models[model_name].load_state_dict(
                                    {
                                        key: checkpoint[
                                            "{}_state_dict".format(state_dict_name)
                                        ][key]
                                    },
                                    strict=False,
                                )
                        except RuntimeError as e:
                            # exception when adding linebreak token from pretraining
                            print(e, flush=True)
                # Set parameters no trainable
                if not learnable:
                    self.set_model_learnable(self.models[model_name], False)

    def adapt_decision_layer_to_old_charset(
        self, model_name, key, checkpoint, state_dict_name
    ):
        """
        Transfer learning of the decision learning in case of close charsets between pre-training and training
        """
        pretrained_chars = list()
        weights = checkpoint["{}_state_dict".format(state_dict_name)][key]
        new_size = list(weights.size())
        new_size[0] = (
            len(self.dataset.charset) + self.params["model"]["additional_tokens"]
        )
        new_weights = torch.zeros(new_size, device=weights.device, dtype=weights.dtype)
        old_charset = (
            checkpoint["charset"]
            if "charset" in checkpoint
            else self.params["model"]["old_charset"]
        )
        if "bias" not in key:
            kaiming_uniform_(new_weights, nonlinearity="relu")
        for i, c in enumerate(self.dataset.charset):
            if c in old_charset:
                new_weights[i] = weights[old_charset.index(c)]
                pretrained_chars.append(c)
        checkpoint["{}_state_dict".format(state_dict_name)][key] = new_weights
        self.models[model_name].load_state_dict(
            {key: checkpoint["{}_state_dict".format(state_dict_name)][key]},
            strict=False,
        )
        print(
            "Pretrained chars for {} ({}): {}".format(
                key, len(pretrained_chars), pretrained_chars
            )
        )

    def load_optimizers(self, checkpoint, reset_optimizer=False):
        """
        Load the optimizer of each model
        """
        for model_name in self.models:
            if (
                checkpoint
                and "optimizer_named_params_{}".format(model_name) in checkpoint
            ):
                self.optimizers_named_params_by_group[model_name] = checkpoint[
                    "optimizer_named_params_{}".format(model_name)
                ]
            else:
                self.optimizers_named_params_by_group[model_name] = [
                    dict(),
                ]
                self.optimizers_named_params_by_group[model_name][0].update(
                    self.models[model_name].named_parameters()
                )

            # Instantiate optimizer
            self.reset_optimizer(model_name)

            # Handle learning rate schedulers
            if self.params["training"].get("lr_schedulers"):
                key = (
                    "all"
                    if "all" in self.params["training"]["lr_schedulers"]
                    else model_name
                )
                if key in self.params["training"]["lr_schedulers"]:
                    self.lr_schedulers[model_name] = self.params["training"][
                        "lr_schedulers"
                    ][key]["class"](
                        self.optimizers[model_name],
                        **self.params["training"]["lr_schedulers"][key]["args"],
                    )

            # Load optimizer state from past training
            if checkpoint and not reset_optimizer:
                self.optimizers[model_name].load_state_dict(
                    checkpoint["optimizer_{}_state_dict".format(model_name)]
                )
                # Load optimizer scheduler config from past training if used
                if (
                    "lr_schedulers" in self.params["training"]
                    and self.params["training"]["lr_schedulers"]
                    and "lr_scheduler_{}_state_dict".format(model_name) in checkpoint
                ):
                    self.lr_schedulers[model_name].load_state_dict(
                        checkpoint["lr_scheduler_{}_state_dict".format(model_name)]
                    )

    @staticmethod
    def set_model_learnable(model, learnable=True):
        for p in list(model.parameters()):
            p.requires_grad = learnable

    def save_model(self, epoch: int, name: str):
        """
        Save model weights and training info for curriculum learning or learning rate for instance
        """
        if not self.is_master:
            return

        path = self.paths["checkpoints"] / "{}_{}.pt".format(name, epoch)
        content = {
            "optimizers_named_params": self.optimizers_named_params_by_group,
            "epoch": epoch,
            "step": self.latest_step,
            "scaler_state_dict": self.scaler.state_dict(),
            "best": self.best,
            "charset": self.dataset.charset,
            "dropout_scheduler_step": self.dropout_scheduler.step_num,
        }

        for model_name in self.optimizers:
            content["optimizer_{}_state_dict".format(model_name)] = self.optimizers[
                model_name
            ].state_dict()

        for model_name in self.lr_schedulers:
            content[
                "lr_scheduler_{}_state_dict".format(model_name)
            ] = self.lr_schedulers[model_name].state_dict()

        content = self.add_save_info(content)
        for model_name in self.models:
            content["{}_state_dict".format(model_name)] = self.models[
                model_name
            ].state_dict()

        # Remove other checkpoints
        for path_to_del in self.paths["checkpoints"].glob(f"{name}_*.pt"):
            logger.warning(f"Removing checkpoint `{path_to_del}`")
            path_to_del.unlink()

        torch.save(content, path)

    def reset_optimizer(self, model_name):
        """
        Reset optimizer learning rate for given model
        """
        params = list(self.optimizers_named_params_by_group[model_name][0].values())
        key = "all" if "all" in self.params["training"]["optimizers"] else model_name
        self.optimizers[model_name] = self.params["training"]["optimizers"][key][
            "class"
        ](params, **self.params["training"]["optimizers"][key]["args"])
        for i in range(1, len(self.optimizers_named_params_by_group[model_name])):
            self.optimizers[model_name].add_param_group(
                {
                    "params": list(
                        self.optimizers_named_params_by_group[model_name][i].values()
                    )
                }
            )

    def save_params(self):
        """
        Output a yaml file containing a summary of all hyperparameters chosen for the training
        and a yaml file containing parameters used for inference
        """

        def compute_nb_params(module) -> np.int64:
            return sum([np.prod(p.size()) for p in list(module.parameters())])

        def class_to_str_dict(my_dict):
            for key in my_dict:
                if key == "preprocessings":
                    my_dict[key] = [
                        {
                            key: value.value if isinstance(value, Enum) else value
                            for key, value in preprocessing.items()
                        }
                        for preprocessing in my_dict[key]
                    ]
                elif callable(my_dict[key]):
                    my_dict[key] = my_dict[key].__name__
                elif isinstance(my_dict[key], np.ndarray):
                    my_dict[key] = my_dict[key].tolist()
                elif isinstance(my_dict[key], list) and isinstance(
                    my_dict[key][0], tuple
                ):
                    my_dict[key] = [list(elt) for elt in my_dict[key]]
                elif isinstance(my_dict[key], Path):
                    my_dict[key] = str(my_dict[key])
                elif isinstance(my_dict[key], dict):
                    my_dict[key] = class_to_str_dict(my_dict[key])
            return my_dict

        # Save training parameters
        path = self.paths["results"] / "training_parameters.yml"
        if path.is_file():
            return

        params = class_to_str_dict(my_dict=deepcopy(self.params))

        # Special case for transfer_learning parameter serialization
        if params["training"].get("transfer_learning"):
            for data in params["training"]["transfer_learning"].values():
                # The second item is a pathlib.Path object
                data[1] = str(data[1])

        total_params = 0
        for model_name in MODEL_NAMES:
            current_params = int(compute_nb_params(self.models[model_name]))
            params["model"][model_name]["nb_params"] = current_params
            total_params += current_params
        params["model"]["total_params"] = "{:,}".format(total_params)
        params["mean"] = self.dataset.mean.tolist()
        params["std"] = self.dataset.std.tolist()

        path.write_text(yaml.safe_dump(params))

        # Save inference parameters
        path = self.paths["results"] / "inference_parameters.yml"
        if path.is_file():
            return

        decoder_params = {
            key: params["model"]["decoder"][key]
            for key in (
                "l_max",
                "dec_num_layers",
                "dec_num_heads",
                "dec_res_dropout",
                "dec_pred_dropout",
                "dec_att_dropout",
                "dec_dim_feedforward",
                "attention_win",
                "enc_dim",
            )
        }

        inference_params = {
            "parameters": {
                "mean": params["mean"],
                "std": params["std"],
                "max_char_prediction": params["dataset"]["max_char_prediction"],
                "encoder": {"dropout": params["model"]["encoder"]["dropout"]},
                "decoder": {
                    "h_max": params["model"]["h_max"],
                    "w_max": params["model"]["w_max"],
                    "vocab_size": params["model"]["vocab_size"],
                    **decoder_params,
                },
                "preprocessings": params["training"]["data"]["preprocessings"],
            },
        }
        path.write_text(yaml.safe_dump(inference_params))

    def backward_loss(self, loss, retain_graph=False):
        self.scaler.scale(loss).backward(retain_graph=retain_graph)

    def step_optimizers(self, names=None):
        for model_name in self.optimizers:
            if names and model_name not in names:
                continue
            if (
                self.params["training"].get("gradient_clipping")
                and model_name in self.params["training"]["gradient_clipping"]["models"]
            ):
                self.scaler.unscale_(self.optimizers[model_name])
                torch.nn.utils.clip_grad_norm_(
                    self.models[model_name].parameters(),
                    self.params["training"]["gradient_clipping"]["max"],
                )
            self.scaler.step(self.optimizers[model_name])
        self.scaler.update()
        self.latest_step += 1

    def zero_optimizers(self, set_to_none=True):
        for model_name in self.optimizers:
            self.optimizers[model_name].zero_grad(set_to_none=set_to_none)

    def train(self, mlflow_logging=False):
        """
        Main training loop
        """
        # init tensorboard file and output param summary file
        if self.is_master:
            self.writer = SummaryWriter(self.paths["results"])
            self.save_params()

        # init variables
        nb_epochs = self.params["training"]["max_nb_epochs"]
        metric_names = self.params["training"]["metrics"]["train"]

        display_values = None
        # perform epochs
        for num_epoch in range(self.latest_epoch + 1, nb_epochs):
            # set models trainable
            for model_name in self.models:
                self.models[model_name].train()
            self.latest_epoch = num_epoch
            if self.dataset.train_dataset.curriculum_config:
                self.dataset.train_dataset.curriculum_config[
                    "epoch"
                ] = self.latest_epoch
            # init epoch metrics values
            self.metric_manager["train"] = MetricManager(
                metric_names=metric_names,
                dataset_name=self.dataset_name,
                tokens=self.tokens,
            )
            with tqdm(total=len(self.dataset.train_loader.dataset)) as pbar:
                pbar.set_description("EPOCH {}/{}".format(num_epoch, nb_epochs))
                # iterates over mini-batch data
                for ind_batch, batch_data in enumerate(self.dataset.train_loader):
                    # train on batch data and compute metrics
                    batch_values = self.train_batch(batch_data, metric_names)
                    batch_metrics = self.metric_manager["train"].compute_metrics(
                        batch_values, metric_names
                    )
                    batch_metrics["names"] = batch_data["names"]
                    # Merge metrics if Distributed Data Parallel is used
                    if self.device_params["use_ddp"]:
                        batch_metrics = self.merge_ddp_metrics(batch_metrics)
                    # Update learning rate via scheduler if one is used
                    if self.params["training"]["lr_schedulers"]:
                        for model_name in self.models:
                            key = (
                                "all"
                                if "all" in self.params["training"]["lr_schedulers"]
                                else model_name
                            )
                            if (
                                model_name in self.lr_schedulers
                                and ind_batch
                                % self.params["training"]["lr_schedulers"][key][
                                    "step_interval"
                                ]
                                == 0
                            ):
                                self.lr_schedulers[model_name].step(
                                    len(batch_metrics["names"])
                                )
                                if "lr" in metric_names:
                                    self.writer.add_scalar(
                                        "lr/{}".format(model_name),
                                    )

                    # Update dropout scheduler
                    self.dropout_scheduler.step(len(batch_metrics["names"]))
                    self.dropout_scheduler.update_dropout_rate()

                    # Add batch metrics values to epoch metrics values
                    self.metric_manager["train"].update_metrics(batch_metrics)
                    display_values = self.metric_manager["train"].get_display_values()
                    pbar.set_postfix(values=str(display_values))
                    pbar.update(len(batch_data["names"]) * self.nb_workers)

                # Log MLflow metrics
                logging_metrics(
                    display_values, "train", num_epoch, mlflow_logging, self.is_master
                )

            if self.is_master:
                # log metrics in tensorboard file
                for key in display_values:
                    self.writer.add_scalar(
                        "train/{}_{}".format(
                            self.params["dataset"]["train"]["name"], key
                        ),
                        display_values[key],
                        num_epoch,
                    )

            # evaluate and compute metrics for valid sets
            if (
                self.params["training"]["validation"]["eval_on_valid"]
                and num_epoch
                % self.params["training"]["validation"]["eval_on_valid_interval"]
                == 0
            ):
                for valid_set_name in self.dataset.valid_loaders:
                    # evaluate set and compute metrics
                    eval_values = self.validate(
                        valid_set_name, mlflow_logging=mlflow_logging
                    )
                    # log valid metrics in tensorboard file
                    if self.is_master:
                        for key in eval_values:
                            self.writer.add_scalar(
                                "valid/{}_{}".format(valid_set_name, key),
                                eval_values[key],
                                num_epoch,
                            )
                        if valid_set_name == self.params["training"]["validation"][
                            "set_name_focus_metric"
                        ] and (self.best is None or eval_values["cer"] <= self.best):
                            self.save_model(epoch=num_epoch, name="best")
                            self.best = eval_values["cer"]

            # save model weights
            if self.is_master:
                self.save_model(epoch=num_epoch, name="last")
                self.writer.flush()

    def validate(self, set_name, mlflow_logging=False, **kwargs):
        """
        Main loop for validation
        """
        loader = self.dataset.valid_loaders[set_name]
        # Set models in eval mode
        for model_name in self.models:
            self.models[model_name].eval()
        metric_names = self.params["training"]["metrics"]["eval"]
        display_values = None

        # initialize epoch metrics
        self.metric_manager[set_name] = MetricManager(
            metric_names=metric_names,
            dataset_name=self.dataset_name,
            tokens=self.tokens,
        )
        with tqdm(total=len(loader.dataset)) as pbar:
            pbar.set_description("VALID {} - {}".format(self.latest_epoch, set_name))
            with torch.no_grad():
                # iterate over batch data
                for ind_batch, batch_data in enumerate(loader):
                    # eval batch data and compute metrics
                    batch_values = self.evaluate_batch(batch_data, metric_names)
                    batch_metrics = self.metric_manager[set_name].compute_metrics(
                        batch_values, metric_names
                    )
                    batch_metrics["names"] = batch_data["names"]
                    # merge metrics values if Distributed Data Parallel is used
                    if self.device_params["use_ddp"]:
                        batch_metrics = self.merge_ddp_metrics(batch_metrics)

                    # add batch metrics to epoch metrics
                    self.metric_manager[set_name].update_metrics(batch_metrics)
                    display_values = self.metric_manager[set_name].get_display_values()

                    pbar.set_postfix(values=str(display_values))
                    pbar.update(len(batch_data["names"]) * self.nb_workers)

                # log metrics in MLflow
                logging_metrics(
                    display_values,
                    "val",
                    self.latest_epoch,
                    mlflow_logging,
                    self.is_master,
                )
        return display_values

    def evaluate(
        self, custom_name, sets_list, metric_names, mlflow_logging=False
    ) -> Tuple[Dict[str, int | float], List[Inference]]:
        """
        Main loop for evaluation
        """
        metric_names = metric_names.copy()
        self.dataset.generate_test_loader(custom_name, sets_list)
        loader = self.dataset.test_loaders[custom_name]
        # Set models in eval mode
        for model_name in self.models:
            self.models[model_name].eval()

        # initialize epoch metrics
        self.metric_manager[custom_name] = MetricManager(
            metric_names=metric_names,
            dataset_name=self.dataset_name,
            tokens=self.tokens,
        )

        # Keep inferences in memory to:
        # - evaluate with Nerval
        # - display worst predictions
        inferences = []

        with tqdm(total=len(loader.dataset)) as pbar:
            pbar.set_description("Evaluation")
            with torch.no_grad():
                for ind_batch, batch_data in enumerate(loader):
                    # iterates over batch data
                    # eval batch data and compute metrics
                    batch_values = self.evaluate_batch(batch_data, metric_names)
                    batch_metrics = self.metric_manager[custom_name].compute_metrics(
                        batch_values, metric_names
                    )
                    batch_metrics["names"] = batch_data["names"]
                    # merge batch metrics if Distributed Data Parallel is used
                    if self.device_params["use_ddp"]:
                        batch_metrics = self.merge_ddp_metrics(batch_metrics)

                    # add batch metrics to epoch metrics
                    self.metric_manager[custom_name].update_metrics(batch_metrics)
                    display_values = self.metric_manager[
                        custom_name
                    ].get_display_values()

                    pbar.set_postfix(values=str(display_values))
                    pbar.update(len(batch_data["names"]) * self.nb_workers)
                    inferences.extend(
                        map(
                            Inference,
                            batch_data["names"],
                            batch_values["str_y"],
                            batch_values["str_x"],
                            batch_values.get("str_lm", repeat("")),
                            batch_metrics["words_error_rate"],
                        )
                    )

                # log metrics in MLflow
                logging_name = custom_name.split("-")[1]
                logging_tags_metrics(
                    display_values, logging_name, mlflow_logging, self.is_master
                )

        if "pred" in metric_names:
            self.output_pred(custom_name)
        metrics = self.metric_manager[custom_name].get_display_values(output=True)
        path = self.paths["results"] / "predict_{}_{}.yaml".format(
            custom_name, self.latest_epoch
        )
        path.write_text(yaml.dump(metrics))

        if mlflow_logging:
            # Log mlflow artifacts
            mlflow.log_artifact(path, "predictions")

        return metrics, inferences

    def output_pred(self, name):
        path = self.paths["results"] / "predict_{}_{}.yaml".format(
            name, self.latest_epoch
        )

        pred = "\n".join(self.metric_manager[name].get("pred"))
        path.write_text(yaml.dump(pred))

    def launch_ddp(self):
        """
        Initialize Distributed Data Parallel system
        """
        mp.set_start_method("fork", force=True)
        os.environ["MASTER_ADDR"] = self.ddp_config["address"]
        os.environ["MASTER_PORT"] = str(self.ddp_config["port"])
        dist.init_process_group(
            self.ddp_config["backend"],
            rank=self.ddp_config["rank"],
            world_size=self.nb_gpu,
        )
        torch.cuda.set_device(self.ddp_config["rank"])
        random.seed(self.manual_seed)
        np.random.seed(self.manual_seed)
        torch.manual_seed(self.manual_seed)
        torch.cuda.manual_seed(self.manual_seed)

    def merge_ddp_metrics(self, metrics):
        """
        Merge metrics when Distributed Data Parallel is used
        """
        for metric_name in metrics:
            if metric_name in [
                "edit_words",
                "nb_words",
                "edit_chars",
                "nb_chars",
                "edit_chars_force_len",
                "edit_chars_curr",
                "nb_chars_curr",
            ]:
                metrics[metric_name] = self.cat_ddp_metric(metrics[metric_name])
            elif metric_name in [
                "nb_samples",
                "loss",
                "loss_ce",
                "loss_ce_end",
            ]:
                metrics[metric_name] = self.sum_ddp_metric(
                    metrics[metric_name], average=False
                )
        return metrics

    def sum_ddp_metric(self, metric, average=False):
        """
        Sum metrics for Distributed Data Parallel
        """
        sum = torch.tensor(metric[0]).to(self.device)
        dist.all_reduce(sum, op=dist.ReduceOp.SUM)
        if average:
            sum.true_divide(dist.get_world_size())
        return [
            sum.item(),
        ]

    def cat_ddp_metric(self, metric):
        """
        Concatenate metrics for Distributed Data Parallel
        """
        tensor = torch.tensor(metric).unsqueeze(0).to(self.device)
        res = [
            torch.zeros(tensor.size()).long().to(self.device)
            for _ in range(dist.get_world_size())
        ]
        dist.all_gather(res, tensor)
        return list(torch.cat(res, dim=0).flatten().cpu().numpy())

    def train_batch(self, batch_data, metric_names):
        raise NotImplementedError

    def evaluate_batch(self, batch_data, metric_names):
        raise NotImplementedError

    def load_save_info(self, info_dict):
        """
        Load curriculum info from saved model info
        """
        if "curriculum_config" in info_dict:
            self.dataset.train_dataset.curriculum_config = info_dict[
                "curriculum_config"
            ]

    def add_save_info(self, info_dict):
        """
        Add curriculum info to model info to be saved
        """
        info_dict["curriculum_config"] = self.dataset.train_dataset.curriculum_config
        return info_dict


class Manager(GenericTrainingManager):
    def __init__(self, params):
        super(Manager, self).__init__(params)
        self.params["model"]["vocab_size"] = len(self.dataset.charset)

    def load_save_info(self, info_dict):
        if "curriculum_config" in info_dict:
            if self.dataset.train_dataset is not None:
                self.dataset.train_dataset.curriculum_config = info_dict[
                    "curriculum_config"
                ]

    def add_save_info(self, info_dict):
        info_dict["curriculum_config"] = self.dataset.train_dataset.curriculum_config
        return info_dict

    def add_label_noise(self, y, y_len, error_rate):
        y_error = y.clone()
        for b in range(len(y_len)):
            for i in range(1, y_len[b]):
                if (
                    np.random.rand() < error_rate
                    and y[b][i] != self.dataset.tokens["pad"]
                ):
                    y_error[b][i] = np.random.randint(0, len(self.dataset.charset) + 2)
        return y_error, y_len

    def train_batch(self, batch_data, metric_names):
        loss_func = CrossEntropyLoss(ignore_index=self.dataset.tokens["pad"])

        sum_loss = 0
        b = batch_data["imgs"].shape[0]
        batch_data["labels"] = batch_data["labels"].to(self.device)
        y_len = batch_data["labels_len"]

        if "label_noise_scheduler" in self.params["training"]:
            error_rate = (
                self.params["training"]["label_noise_scheduler"]["min_error_rate"]
                + min(
                    self.latest_step,
                    self.params["training"]["label_noise_scheduler"]["total_num_steps"],
                )
                * (
                    self.params["training"]["label_noise_scheduler"]["max_error_rate"]
                    - self.params["training"]["label_noise_scheduler"]["min_error_rate"]
                )
                / self.params["training"]["label_noise_scheduler"]["total_num_steps"]
            )
            simulated_y_pred, y_len = self.add_label_noise(
                batch_data["labels"], y_len, error_rate
            )
        else:
            simulated_y_pred = batch_data["labels"]

        with autocast(enabled=self.device_params["use_amp"]):
            hidden_predict = None
            cache = None

            features = self.encoder(batch_data["imgs"].to(self.device))
            features_size = features.size()

            if self.device_params["use_ddp"]:
                features = self.decoder.module.features_updater.get_pos_features(
                    features
                )
            else:
                features = self.decoder.features_updater.get_pos_features(features)
            features = torch.flatten(features, start_dim=2, end_dim=3).permute(2, 0, 1)

            output, pred, hidden_predict, cache, weights = self.decoder(
                features,
                simulated_y_pred[:, :-1],
                [s[:2] for s in batch_data["imgs_reduced_shape"]],
                [max(y_len) for _ in range(b)],
                features_size,
                start=0,
                hidden_predict=hidden_predict,
                cache=cache,
                keep_all_weights=True,
            )

            loss_ce = loss_func(pred, batch_data["labels"][:, 1:])
            sum_loss += loss_ce
            with autocast(enabled=False):
                self.backward_loss(sum_loss)
                self.step_optimizers()
                self.zero_optimizers()
            predicted_tokens = torch.argmax(pred, dim=1).detach().cpu().numpy()
            predicted_tokens = [predicted_tokens[i, : y_len[i]] for i in range(b)]
            str_x = [
                ind_to_token(self.dataset.charset, t, oov_symbol="")
                for t in predicted_tokens
            ]

        values = {
            "nb_samples": b,
            "str_y": batch_data["raw_labels"],
            "str_x": str_x,
            "loss": sum_loss.item(),
            "loss_ce": loss_ce.item(),
        }

        return values

    def evaluate_batch(self, batch_data, metric_names):
        x = batch_data["imgs"].to(self.device)

        max_chars = self.params["dataset"]["max_char_prediction"]

        start_time = time()
        with autocast(enabled=self.device_params["use_amp"]):
            b = x.size(0)
            reached_end = torch.zeros((b,), dtype=torch.bool, device=self.device)
            prediction_len = torch.zeros((b,), dtype=torch.int, device=self.device)
            predicted_tokens = (
                torch.ones((b, 1), dtype=torch.long, device=self.device)
                * self.dataset.tokens["start"]
            )
            predicted_tokens_len = torch.ones((b,), dtype=torch.int, device=self.device)

            # end token index will be used for ctc
            tot_pred = torch.zeros(
                (b, len(self.dataset.charset) + 1, max_chars),
                dtype=torch.float,
                device=self.device,
            )

            whole_output = list()
            confidence_scores = list()
            cache = None
            hidden_predict = None
            if b > 1:
                features_list = list()
                for i in range(b):
                    pos = batch_data["imgs_position"]
                    features_list.append(
                        self.encoder(
                            x[
                                i : i + 1,
                                :,
                                pos[i][0][0] : pos[i][0][1],
                                pos[i][1][0] : pos[i][1][1],
                            ]
                        )
                    )
                max_height = max([f.size(2) for f in features_list])
                max_width = max([f.size(3) for f in features_list])
                features = torch.zeros(
                    (b, features_list[0].size(1), max_height, max_width),
                    device=self.device,
                    dtype=features_list[0].dtype,
                )
                for i in range(b):
                    features[
                        i, :, : features_list[i].size(2), : features_list[i].size(3)
                    ] = features_list[i]
            else:
                features = self.encoder(x)
            features_size = features.size()

            if self.device_params["use_ddp"]:
                features = self.decoder.module.features_updater.get_pos_features(
                    features
                )
            else:
                features = self.decoder.features_updater.get_pos_features(features)
            features = torch.flatten(features, start_dim=2, end_dim=3).permute(2, 0, 1)

            for i in range(0, max_chars):
                output, pred, hidden_predict, cache, weights = self.decoder(
                    features,
                    predicted_tokens,
                    [s[:2] for s in batch_data["imgs_reduced_shape"]],
                    predicted_tokens_len,
                    features_size,
                    start=0,
                    hidden_predict=hidden_predict,
                    cache=cache,
                    num_pred=1,
                )

                # output total logit prediction
                tot_pred[:, :, i : i + 1] = pred

                whole_output.append(output)
                confidence_scores.append(
                    torch.max(torch.softmax(pred[:, :], dim=1), dim=1).values
                )
                predicted_tokens = torch.cat(
                    [
                        predicted_tokens,
                        torch.argmax(pred[:, :, -1], dim=1, keepdim=True),
                    ],
                    dim=1,
                )
                reached_end = torch.logical_or(
                    reached_end,
                    torch.eq(predicted_tokens[:, -1], self.dataset.tokens["end"]),
                )
                predicted_tokens_len += 1

                prediction_len[reached_end == False] = i + 1  # noqa E712
                if torch.all(reached_end):
                    break

            confidence_scores = (
                torch.cat(confidence_scores, dim=1).cpu().detach().numpy()
            )
            predicted_tokens = predicted_tokens[:, 1:]
            prediction_len[torch.eq(reached_end, False)] = max_chars - 1
            predicted_tokens = [
                predicted_tokens[i, : prediction_len[i]] for i in range(b)
            ]
            confidence_scores = [
                confidence_scores[i, : prediction_len[i]].tolist() for i in range(b)
            ]
            str_x = [
                ind_to_token(self.dataset.charset, t, oov_symbol="")
                for t in predicted_tokens
            ]

        process_time = time() - start_time

        values = {
            "nb_samples": b,
            "str_y": batch_data["raw_labels"],
            "str_x": str_x,
            "confidence_score": confidence_scores,
            "time": process_time,
        }
        if self.lm_decoder:
            values["str_lm"] = self.lm_decoder(tot_pred, prediction_len)["text"]

        return values



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import logging
from collections import Counter, defaultdict
from functools import partial
from pathlib import Path
from typing import Dict, List

import imagesize
import numpy as np
from mdutils.mdutils import MdUtils
from prettytable import MARKDOWN, PrettyTable

from dan.utils import read_json

logger = logging.getLogger(__name__)

METRIC_COLUMN = "Metric"


def create_table(
    data: Dict,
    count: bool = False,
    total: bool = True,
):
    """
    Each keys will be made into a column
    We compute min, max, mean, median, total by default.
    Total can be disabled. Count (length) computation can be enabled.
    """

    statistics = PrettyTable(field_names=[METRIC_COLUMN, *data.keys()])
    statistics.align.update({METRIC_COLUMN: "l"})
    statistics.set_style(MARKDOWN)

    operations = []

    if count:
        operations.append(("Count", len, None))

    operations.extend(
        [
            ("Min", np.min, None),
            ("Max", np.max, None),
            ("Mean", np.mean, 2),
            ("Median", np.median, 2),
        ]
    )
    if total:
        operations.append(("Total", np.sum, None))

    statistics.add_rows(
        [
            [
                col_name,
                *list(
                    map(
                        # Round values if needed
                        partial(round, ndigits=digits),
                        map(operator, data.values()),
                    )
                ),
            ]
            for col_name, operator, digits in operations
        ]
    )

    return statistics


class Statistics:
    HEADERS = {
        "Images": "Images statistics",
        "Labels": "Labels statistics",
        "Chars": "Characters statistics",
        "Tokens": "NER tokens statistics",
    }

    def __init__(self, filename: str) -> None:
        self.document = MdUtils(file_name=filename, title="Statistics")

    def _write_section(self, table: PrettyTable, title: str, level: int = 2):
        """
        Write the new section in the file.

        <title with appropriate level>

        <table>

        """
        self.document.new_header(level=level, title=title, add_table_of_contents="n")
        self.document.write("\n")

        logger.info(f"{title}\n\n{table}\n")

        self.document.write(table.get_string())
        self.document.write("\n")

    def create_image_statistics(self, images: List[str]):
        """
        Compute statistics on image sizes and write them to file.
        """
        shapes = list(map(imagesize.get, images))
        widths, heights = zip(*shapes)

        self._write_section(
            table=create_table(
                data={"Width": widths, "Height": heights}, count=True, total=False
            ),
            title=Statistics.HEADERS["Images"],
        )

    def create_label_statistics(self, labels: List[str]):
        """
        Compute statistics on text labels and write them to file.
        """
        char_counter = Counter()
        data = defaultdict(list)

        for text in labels:
            char_counter.update(text)
            data["Chars"].append(len(text))
            data["Words"].append(len(text.split()))
            data["Lines"].append(len(text.split("\n")))

        self._write_section(
            table=create_table(data=data),
            title=Statistics.HEADERS["Labels"],
        )

        self.create_character_occurrences_statistics(char_counter)

    def create_character_occurrences_statistics(self, char_counter: Counter):
        """
        Compute statistics on the character distribution and write them to file.
        """
        char_occurrences = PrettyTable(
            field_names=["Character", "Occurrence"],
        )
        char_occurrences.align.update({"Character": "l", "Occurrence": "r"})
        char_occurrences.set_style(MARKDOWN)
        char_occurrences.add_rows(list(char_counter.most_common()))

        self._write_section(
            table=char_occurrences, title=Statistics.HEADERS["Chars"], level=3
        )

    def create_ner_statistics(self, labels: List[str], ner_tokens: Dict) -> str:
        """
        Compute statistics on ner tokens presence.
        """
        entity_counter = defaultdict(list)
        for text in labels:
            for ner_label, token in ner_tokens.items():
                entity_counter[ner_label].append(text.count(token["start"]))

        self._write_section(
            table=create_table(data=entity_counter),
            title=Statistics.HEADERS["Tokens"],
            level=3,
        )

    def run(self, labels_path: Path, tokens: Dict | None):
        labels = read_json(labels_path)

        # Iterate over each split
        for split_name, split_data in labels.items():
            self.document.new_header(level=1, title=split_name.capitalize())

            # Image statistics
            # Path to the images are the key of the dict
            self.create_image_statistics(
                images=[labels_path.parent / image_path for image_path in split_data]
            )

            labels = list(split_data.values())
            # Text statistics
            self.create_label_statistics(labels=labels)

            if tokens is not None:
                self.create_ner_statistics(labels=labels, ner_tokens=tokens)
        self.document.create_md_file()


def run(labels: Path, tokens: Dict | None, output: Path) -> None:
    """
    Compute and save a dataset statistics.
    """
    Statistics(filename=str(output)).run(labels_path=labels, tokens=tokens)
